@@ -16,6 +16,7 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  *
  * @author Andre Moreira Magalhaes <andre.magalhaes@openbossa.org>
+ * @author Gustavo Sverzut Barbieri <gustavo.barbieri@openbossa.org>
  */
 
 /**
@@ -45,168 +46,18 @@
 #define ID3V2_HEADER_SIZE       10
 #define ID3V2_FOOTER_SIZE       10
 
-#define ID3V1_NUM_GENRES 148
-
-#define ID3_NUM_ENCODINGS 5
-
 enum ID3Encodings {
     ID3_ENCODING_LATIN1 = 0,
     ID3_ENCODING_UTF16,
     ID3_ENCODING_UTF16BE,
     ID3_ENCODING_UTF8,
-    ID3_ENCODING_UTF16LE
+    ID3_ENCODING_UTF16LE,
+    ID3_ENCODING_LAST
 };
+#define ID3_NUM_ENCODINGS ID3_ENCODING_LAST
 
-static const char *id3v1_genres[ID3V1_NUM_GENRES] = {
-    "Blues",
-    "Classic Rock",
-    "Country",
-    "Dance",
-    "Disco",
-    "Funk",
-    "Grunge",
-    "Hip-Hop",
-    "Jazz",
-    "Metal",
-    "New Age",
-    "Oldies",
-    "Other",
-    "Pop",
-    "R&B",
-    "Rap",
-    "Reggae",
-    "Rock",
-    "Techno",
-    "Industrial",
-    "Alternative",
-    "Ska",
-    "Death Metal",
-    "Pranks",
-    "Soundtrack",
-    "Euro-Techno",
-    "Ambient",
-    "Trip-Hop",
-    "Vocal",
-    "Jazz+Funk",
-    "Fusion",
-    "Trance",
-    "Classical",
-    "Instrumental",
-    "Acid",
-    "House",
-    "Game",
-    "Sound Clip",
-    "Gospel",
-    "Noise",
-    "Alternative Rock",
-    "Bass",
-    "Soul",
-    "Punk",
-    "Space",
-    "Meditative",
-    "Instrumental Pop",
-    "Instrumental Rock",
-    "Ethnic",
-    "Gothic",
-    "Darkwave",
-    "Techno-Industrial",
-    "Electronic",
-    "Pop-Folk",
-    "Eurodance",
-    "Dream",
-    "Southern Rock",
-    "Comedy",
-    "Cult",
-    "Gangsta",
-    "Top 40",
-    "Christian Rap",
-    "Pop/Funk",
-    "Jungle",
-    "Native American",
-    "Cabaret",
-    "New Wave",
-    "Psychedelic",
-    "Rave",
-    "Showtunes",
-    "Trailer",
-    "Lo-Fi",
-    "Tribal",
-    "Acid Punk",
-    "Acid Jazz",
-    "Polka",
-    "Retro",
-    "Musical",
-    "Rock & Roll",
-    "Hard Rock",
-    "Folk",
-    "Folk/Rock",
-    "National Folk",
-    "Swing",
-    "Fusion",
-    "Bebob",
-    "Latin",
-    "Revival",
-    "Celtic",
-    "Bluegrass",
-    "Avantgarde",
-    "Gothic Rock",
-    "Progressive Rock",
-    "Psychedelic Rock",
-    "Symphonic Rock",
-    "Slow Rock",
-    "Big Band",
-    "Chorus",
-    "Easy Listening",
-    "Acoustic",
-    "Humour",
-    "Speech",
-    "Chanson",
-    "Opera",
-    "Chamber Music",
-    "Sonata",
-    "Symphony",
-    "Booty Bass",
-    "Primus",
-    "Porn Groove",
-    "Satire",
-    "Slow Jam",
-    "Club",
-    "Tango",
-    "Samba",
-    "Folklore",
-    "Ballad",
-    "Power Ballad",
-    "Rhythmic Soul",
-    "Freestyle",
-    "Duet",
-    "Punk Rock",
-    "Drum Solo",
-    "A Cappella",
-    "Euro-House",
-    "Dance Hall",
-    "Goa",
-    "Drum & Bass",
-    "Club-House",
-    "Hardcore",
-    "Terror",
-    "Indie",
-    "BritPop",
-    "Negerpunk",
-    "Polsk Punk",
-    "Beat",
-    "Christian Gangsta Rap",
-    "Heavy Metal",
-    "Black Metal",
-    "Crossover",
-    "Contemporary Christian",
-    "Christian Rock",
-    "Merengue",
-    "Salsa",
-    "Thrash Metal",
-    "Anime",
-    "Jpop",
-    "Synthpop"
-};
+
+#include "id3v1_genres.c"
 
 struct id3_info {
     struct lms_string_size title;
@@ -272,102 +123,89 @@ _is_id3v2_second_synch_byte(unsigned char byte)
 static long
 _find_id3v2(int fd)
 {
-    long buffer_offset = 0;
-    char buffer[3], *p;
-    int buffer_size = sizeof(buffer);
-    const char pattern[] = "ID3";
-    ssize_t nread;
+    static const char pattern[3] = "ID3";
+    char buffer[3];
+    unsigned int prev_part_match, prev_part_match_sync = 0;
+    long buffer_offset;
 
-    /* These variables are used to keep track of a partial match that happens at
-     * the end of a buffer. */
-    int previous_partial_match = -1;
-    int previous_partial_synch_match = 0;
-    int first_synch_byte;
-
-    /* Start the search at the beginning of the file. */
-    lseek(fd, 0, SEEK_SET);
-
-    if ((nread = read(fd, &buffer, buffer_size)) != buffer_size)
+    if (read(fd, buffer, sizeof(buffer)) != sizeof(buffer))
         return -1;
 
-    /* check if pattern is in the beggining of the file */
-    if (memcmp(buffer, pattern, 3) == 0)
+    if (memcmp(buffer, pattern, sizeof(pattern)) == 0)
         return 0;
 
     /* This loop is the crux of the find method.  There are three cases that we
      * want to account for:
-     * (1) The previously searched buffer contained a partial match of the search
-     * pattern and we want to see if the next one starts with the remainder of
-     * that pattern.
+     * (1) The previously searched buffer contained a partial match of the
+     * search pattern and we want to see if the next one starts with the
+     * remainder of that pattern.
      *
      * (2) The search pattern is wholly contained within the current buffer.
      *
      * (3) The current buffer ends with a partial match of the pattern.  We will
      * note this for use in the next iteration, where we will check for the rest
-     * of the pattern. */
+     * of the pattern.
+     */
+    buffer_offset = 0;
+    prev_part_match_sync = 0;
+    prev_part_match = 0;
     while (1) {
+        const char *p, *p_end;
+
         /* (1) previous partial match */
-        if (previous_partial_synch_match && _is_id3v2_second_synch_byte(buffer[0]))
-            return -1;
-
-        if (previous_partial_match >= 0 && previous_partial_match < buffer_size) {
-            const int pattern_offset = buffer_size - previous_partial_match;
-
-            if (memcmp(buffer, pattern + pattern_offset, 3 - pattern_offset) == 0)
-                return buffer_offset - buffer_size + previous_partial_match;
+        if (prev_part_match_sync) {
+            if (_is_id3v2_second_synch_byte(buffer[0]))
+                return -1;
+            prev_part_match_sync = 0;
         }
 
-        /* (2) pattern contained in current buffer */
-        p = buffer;
-        while ((p = memchr(p, 'I', buffer_size - (p - buffer)))) {
-            if (buffer_size - (p - buffer) < 3)
-                break;
+        if (prev_part_match) {
+            const int size = sizeof(buffer) - prev_part_match;
+            const char *part_pattern = pattern + prev_part_match;
 
-            if (memcmp(p, pattern, 3) == 0)
-                return buffer_offset + (p - buffer);
+            if (memcmp(buffer, part_pattern, size) == 0)
+                return buffer_offset - prev_part_match;
 
-            p += 1;
+            prev_part_match = 0;
         }
 
-        p = memchr(buffer, 255, buffer_size);
-        if (p)
-            first_synch_byte = p - buffer;
-        else
-            first_synch_byte = -1;
+        p_end = buffer + sizeof(buffer);
+        for (p = buffer; p < p_end; p++) {
+            if (*p == pattern[0]) {
+                /* Try to match pattern, possible partial contents */
+                const char *q;
+                int todo;
 
-        /* Here we have to loop because there could be several of the first
-         * (11111111) byte, and we want to check all such instances until we find
-         * a full match (11111111 111) or hit the end of the buffer. */
-        while (first_synch_byte >= 0) {
-            /* if this *is not* at the end of the buffer */
-            if (first_synch_byte < buffer_size - 1) {
-                if(_is_id3v2_second_synch_byte(buffer[first_synch_byte + 1]))
-                    /* We've found the frame synch pattern. */
-                    return -1;
-                else
-                    /* We found 11111111 at the end of the current buffer indicating a
-                     * partial match of the synch pattern.  The find() below should
-                     * return -1 and break out of the loop. */
-                    previous_partial_synch_match = 1;
+                q = p + 1;
+                todo = p_end - q;
+                if (todo == 0 || memcmp(q, pattern + 1, todo) == 0) {
+                    todo++;
+                    if (todo == sizeof(buffer))
+                        /* (2) pattern contained in current buffer */
+                        return buffer_offset;
+
+                    /* (3) partial match */
+                    prev_part_match = todo;
+                    break;
+                }
+            } else if ((unsigned char)*p == 0xff) {
+                /* Try to match synch pattern, possible partial contents */
+                const char *q;
+
+                q = p + 1;
+                if (q < p_end) {
+                    if (_is_id3v2_second_synch_byte(*q))
+                        /* (2) synch pattern contained in current buffer */
+                        return -1;
+                } else
+                    /* (3) partial match */
+                    prev_part_match_sync = 1;
             }
-
-            /* Check in the rest of the buffer. */
-            p = memchr(p + 1, 255, buffer_size - (p + 1 - buffer));
-            if (p)
-                first_synch_byte = p - buffer;
-            else
-                first_synch_byte = -1;
         }
 
-        /* (3) partial match */
-        if (buffer[nread - 1] == pattern[1])
-            previous_partial_match = nread - 1;
-        else if (memcmp(&buffer[nread - 2], pattern, 2) == 0)
-            previous_partial_match = nread - 2;
-        buffer_offset += buffer_size;
-
-        if ((nread = read(fd, &buffer, sizeof(buffer))) == -1)
+        if (read(fd, buffer, sizeof(buffer)) != sizeof(buffer))
             return -1;
+        buffer_offset += sizeof(buffer);
     }
 
     return -1;
@@ -418,27 +256,182 @@ _parse_id3v2_frame_header(char *data, unsigned int version, struct id3v2_frame_h
 }
 
 static inline void
-_get_id3v2_frame_info(const char *frame_data, unsigned int frame_size, struct lms_string_size *s, lms_charset_conv_t *cs_conv)
+_get_id3v2_frame_info(const char *frame_data, unsigned int frame_size, struct lms_string_size *s, lms_charset_conv_t *cs_conv, int strip)
 {
-    s->str = malloc(sizeof(char) * (frame_size + 1));
+    if (frame_size == 0 || frame_size == 0)
+        return;
+    if (frame_size > s->len) {
+        char *tmp;
+
+        tmp = realloc(s->str, sizeof(char) * (frame_size + 1));
+        if (!tmp)
+            return;
+        s->str = tmp;
+    }
     memcpy(s->str, frame_data, frame_size);
     s->str[frame_size] = '\0';
     s->len = frame_size;
     if (cs_conv)
         lms_charset_conv(cs_conv, &s->str, &s->len);
+    if (strip)
+        lms_string_size_strip_and_free(s);
+}
+
+static int
+_get_id3v2_artist(unsigned int index, const char *frame_data, unsigned int frame_size, struct id3_info *info, lms_charset_conv_t *cs_conv)
+{
+    static const unsigned char artist_priorities[] = {3, 4, 2, 1};
+    const unsigned int index_max = sizeof(artist_priorities) / sizeof(*artist_priorities);
+
+    if (index >= index_max)
+        return 1;
+
+    if (artist_priorities[index] > info->cur_artist_priority) {
+        struct lms_string_size artist = {0};
+
+        _get_id3v2_frame_info(frame_data, frame_size, &artist, cs_conv, 1);
+        if (artist.str) {
+            if (info->artist.str)
+                free(info->artist.str);
+            info->artist = artist;
+            info->cur_artist_priority = artist_priorities[index];
+        }
+    }
+    return 0;
+}
+
+static int
+_get_id3v1_genre(unsigned int genre, struct lms_string_size *out)
+{
+    if (genre < ID3V1_NUM_GENRES) {
+        unsigned int size, base, len;
+
+        base = id3v1_genres_offsets[genre];
+        size = id3v1_genres_offsets[genre + 1] - base;
+        len = size - 1;
+
+        if (len > out->len) {
+            char *p = realloc(out->str, size);
+            if (!p)
+                return -2;
+            out->str = p;
+        }
+
+        out->len = len;
+        memcpy(out->str, id3v1_genres_mem + base, size);
+
+        return 0;
+    }
+    return -1;
+}
+
+static inline int
+_parse_id3v1_genre(const char *str_genre, struct lms_string_size *out)
+{
+    return _get_id3v1_genre(atoi(str_genre), out);
+}
+
+static void
+_get_id3v2_genre(const char *frame_data, unsigned int frame_size, struct lms_string_size *out, lms_charset_conv_t *cs_conv)
+{
+    int i, is_number;
+    struct lms_string_size genre = {0};
+
+    _get_id3v2_frame_info(frame_data, frame_size, &genre, cs_conv, 1);
+
+    if (!genre.str)
+        return;
+
+    is_number = (genre.len != 0 && genre.str[0] != '(');
+    if (is_number) {
+        for (i = 0; i < genre.len; ++i) {
+            if (!isdigit(genre.str[i])) {
+                is_number = 0;
+                break;
+            }
+        }
+    }
+
+    if (is_number && _parse_id3v1_genre(genre.str, out) == 0) {
+        /* id3v1 genre found */
+        free(genre.str);
+        return;
+    }
+
+    /* ID3v2.3 "content type" can contain a ID3v1 genre number in parenthesis at
+     * the beginning of the field. If this is all that the field contains, do a
+     * translation from that number to the name and return that.  If there is a
+     * string folloing the ID3v1 genre number, that is considered to be
+     * authoritative and we return that instead. Or finally, the field may
+     * simply be free text, in which case we just return the value. */
+
+    if (genre.len > 1 && genre.str[0] == '(') {
+        char *closing = NULL;
+
+        if (genre.str[genre.len - 1] == ')') {
+            closing = strchr(genre.str, ')');
+            if (closing == genre.str + genre.len - 1) {
+                /* ) is the last character and only appears once in the
+                 * string get the id3v1 genre enclosed by parentheses
+                 */
+                if (_parse_id3v1_genre(genre.str + 1, out) == 0) {
+                    free(genre.str);
+                    return;
+                }
+            }
+        }
+
+        /* get the string followed by the id3v1 genre */
+        if (!closing)
+            closing = strchr(genre.str, ')');
+
+        if (closing) {
+            closing++;
+            out->len = genre.len - (closing - genre.str);
+            out->str = genre.str;
+            memmove(out->str, closing, out->len + 1); /* includes '\0' */
+            lms_string_size_strip_and_free(out);
+            return;
+        }
+    }
+
+    /* pure text */
+    *out = genre;
+}
+
+static void
+_get_id3v2_trackno(const char *frame_data, unsigned int frame_size, struct id3_info *info, lms_charset_conv_t *cs_conv)
+{
+    struct lms_string_size trackno = {0};
+
+    _get_id3v2_frame_info(frame_data, frame_size, &trackno, cs_conv, 0);
+    if (!trackno.str)
+        return;
+    info->trackno = atoi(trackno.str);
+    free(trackno.str);
 }
 
 static void
 _parse_id3v2_frame(struct id3v2_frame_header *fh, const char *frame_data, struct id3_info *info, lms_charset_conv_t **cs_convs)
 {
-    lms_charset_conv_t *cs_conv = NULL;
+    lms_charset_conv_t *cs_conv;
     unsigned int text_encoding, frame_size;
-    static const int artist_priorities[] = { 3, 4, 2, 1 };
+    const char *fid;
+
+    /* ignore frames which contains just the encoding */
+    if (fh->frame_size <= 1)
+        return;
 
 #if 0
-    fprintf(stderr, "frame id = %.4s frame size = %d text encoding = %d\n", fh->frame_id, fh->frame_size, frame_data[0]);
+    fprintf(stderr, "frame id = %.4s frame size = %d text encoding = %d\n",
+            fh->frame_id, fh->frame_size, frame_data[0]);
 #endif
 
+    /* All used frames start with 'T' */
+    fid = fh->frame_id;
+    if (fid[0] != 'T')
+        return;
+
     /* Latin1  = 0
      * UTF16   = 1
      * UTF16BE = 2
@@ -451,8 +444,12 @@ _parse_id3v2_frame(struct id3v2_frame_header *fh, const char *frame_data, struct
     frame_data += 1;
     frame_size = fh->frame_size - 1;
 
-    if (text_encoding >= 0 && text_encoding < ID3_NUM_ENCODINGS) {
+    if (text_encoding < ID3_NUM_ENCODINGS) {
         if (text_encoding == ID3_ENCODING_UTF16) {
+            /* ignore frames which contains just the encoding */
+            if (frame_size <= 2)
+                return;
+
             if (memcmp(frame_data, "\xfe\xff", 2) == 0)
                 text_encoding = ID3_ENCODING_UTF16BE;
             else
@@ -461,71 +458,30 @@ _parse_id3v2_frame(struct id3v2_frame_header *fh, const char *frame_data, struct
             frame_size -= 2;
         }
         cs_conv = cs_convs[text_encoding];
-    }
+    } else
+        cs_conv = NULL;
 
     /* ID3v2.2 used 3 bytes for the frame id, so let's check it */
-    if (memcmp(fh->frame_id, "TIT2", 4) == 0 ||
-        memcmp(fh->frame_id, "TT2", 3) == 0)
-        _get_id3v2_frame_info(frame_data, frame_size, &info->title, cs_conv);
-    else if (memcmp(fh->frame_id, "TP", 2) == 0) {
-        int index = -1;
-        if (memcmp(fh->frame_id, "TPE", 3) == 0) {
-            /* this check shouldn't be needed, but let's make sure */
-            if (fh->frame_id[3] >= '1' && fh->frame_id[3] <= '4')
-                index = fh->frame_id[3] - '1';
-        }
-        else {
-            /* ignore TPA, TPB */
-            if (fh->frame_id[2] >= '1' && fh->frame_id[2] <= '4')
-                index = fh->frame_id[2] - '1';
-        }
-
-        if (index != -1 &&
-            artist_priorities[index] > info->cur_artist_priority) {
-            info->cur_artist_priority = artist_priorities[index];
-            _get_id3v2_frame_info(frame_data, frame_size, &info->artist, cs_conv);
-        }
+    if ((fid[1] == 'T' && fid[2] == '2') ||
+        (fid[1] == 'I' && fid[2] == 'T' && fid[3] == '2'))
+        _get_id3v2_frame_info(frame_data, frame_size, &info->title, cs_conv, 1);
+    else if (fid[1] == 'P') {
+        if (fid[2] == 'E')
+            _get_id3v2_artist(fid[3] - '1', frame_data, frame_size,
+                              info, cs_conv);
+        else if (fid[2] >= '1' && fid[2] <= '4')
+            _get_id3v2_artist(fid[2] - '1', frame_data, frame_size,
+                              info, cs_conv);
     }
     /* TALB, TAL */
-    else if (memcmp(fh->frame_id, "TAL", 3) == 0)
-        _get_id3v2_frame_info(frame_data, frame_size, &info->album, cs_conv);
+    else if (fid[1] == 'A' && fid[2] == 'L')
+        _get_id3v2_frame_info(frame_data, frame_size, &info->album, cs_conv, 1);
     /* TCON, TCO */
-    else if (memcmp(fh->frame_id, "TCO", 3) == 0) {
-        /* TODO handle multiple genres */
-        int i, is_number;
-
-        _get_id3v2_frame_info(frame_data, frame_size, &info->genre, cs_conv);
-
-        is_number = 1;
-        for (i = 0; i < info->genre.len; ++i) {
-            if (!isdigit(info->genre.str[i]))
-                is_number = 0;
-        }
-
-        if ((is_number) &&
-            (info->genre.str) &&
-            (info->genre.len > 0)) {
-            int genre = atoi(info->genre.str);
-            if (genre >= 0 && genre < ID3V1_NUM_GENRES) {
-                free(info->genre.str);
-                info->genre.str = strdup(id3v1_genres[(int) genre]);
-                info->genre.len = strlen(info->genre.str);
-            }
-            else {
-                /* ignore other genres */
-                free(info->genre.str);
-                info->genre.str = NULL;
-                info->genre.len = 0;
-            }
-        }
-    }
-    else if (memcmp(fh->frame_id, "TRCK", 4) == 0 ||
-             memcmp(fh->frame_id, "TRK", 3) == 0) {
-        struct lms_string_size trackno;
-        _get_id3v2_frame_info(frame_data, frame_size, &trackno, cs_conv);
-        info->trackno = atoi(trackno.str);
-        free(trackno.str);
-    }
+    else if (fid[1] == 'C' && fid[2] == 'O')
+        _get_id3v2_genre(frame_data, frame_size, &info->genre, cs_conv);
+    else if (fid[1] == 'R' && (fid[2] == 'K' ||
+                               (fid[2] == 'C' && fid[3] == 'K')))
+        _get_id3v2_trackno(frame_data, frame_size, info, cs_conv);
 }
 
 static int
@@ -585,10 +541,9 @@ _parse_id3v2(int fd, long id3v2_offset, struct id3_info *info, lms_charset_conv_
             break;
 
         _parse_id3v2_frame_header(frame_header_data, major_version, &fh);
-        if (!fh.frame_size)
-            break;
 
-        if (!fh.compression &&
+        if (fh.frame_size > 0 &&
+            !fh.compression &&
             fh.frame_id[0] == 'T' &&
             memcmp(fh.frame_id, "TXXX", 4) != 0) {
             char *frame_data;
@@ -618,6 +573,46 @@ _parse_id3v2(int fd, long id3v2_offset, struct id3_info *info, lms_charset_conv_
     return 0;
 }
 
+static inline void
+_id3v1_str_get(struct lms_string_size *s, const char *buf, int maxlen, lms_charset_conv_t *cs_conv)
+{
+    int start, len;
+    const char *p, *p_end, *p_last;
+
+    start = 0;
+    p_last = NULL;
+    p_end = buf + maxlen;
+    for (p = buf; *p != '\0' && p < p_end; p++) {
+        if (!isspace(*p))
+            p_last = p;
+        else if (!p_last)
+            start++;
+    }
+
+    if (!p_last)
+        return;
+
+    len = (p_last - buf) - start;
+    if (len < 1)
+        return;
+
+    if (len > s->len) {
+        char *tmp;
+
+        tmp = realloc(s->str, sizeof(char) * (len + 1));
+        if (!tmp)
+            return;
+        s->str = tmp;
+    }
+
+    s->len = len;
+    memcpy(s->str, buf + start, len);
+    s->str[len] = '\0';
+
+    if (cs_conv)
+        lms_charset_conv(cs_conv, &s->str, &s->len);
+}
+
 static int
 _parse_id3v1(int fd, struct id3_info *info, lms_charset_conv_t *cs_conv)
 {
@@ -625,19 +620,10 @@ _parse_id3v1(int fd, struct id3_info *info, lms_charset_conv_t *cs_conv)
     if (read(fd, &tag, sizeof(struct id3v1_tag)) == -1)
         return -1;
 
-    info->title.str = strndup(tag.title, 30);
-    info->title.len = strlen(info->title.str);
-    lms_charset_conv(cs_conv, &info->title.str, &info->title.len);
-    info->artist.str = strndup(tag.artist, 30);
-    info->artist.len = strlen(info->artist.str);
-    lms_charset_conv(cs_conv, &info->artist.str, &info->artist.len);
-    info->album.str = strndup(tag.album, 30);
-    info->album.len = strlen(info->album.str);
-    lms_charset_conv(cs_conv, &info->album.str, &info->album.len);
-    if ((unsigned char) tag.genre < ID3V1_NUM_GENRES) {
-        info->genre.str = strdup(id3v1_genres[(unsigned char) tag.genre]);
-        info->genre.len = strlen(info->genre.str);
-    }
+    _id3v1_str_get(&info->title, tag.title, sizeof(tag.title), cs_conv);
+    _id3v1_str_get(&info->artist, tag.artist, sizeof(tag.artist), cs_conv);
+    _id3v1_str_get(&info->album, tag.album, sizeof(tag.album), cs_conv);
+    _get_id3v1_genre(tag.genre, &info->genre);
     if (tag.comments[28] == 0 && tag.comments[29] != 0)
         info->trackno = (unsigned char) tag.comments[29];
 
@@ -676,15 +662,15 @@ _parse(struct plugin *plugin, struct lms_context *ctxt, const struct lms_file_in
         fprintf(stderr, "id3v2 tag found in file %s with offset %ld\n",
                 finfo->path, id3v2_offset);
 #endif
-        if (_parse_id3v2(fd, id3v2_offset, &info, plugin->cs_convs) != 0) {
-            r = -2;
-            goto done;
-        }
+        if (_parse_id3v2(fd, id3v2_offset, &info, plugin->cs_convs) != 0)
+            id3v2_offset = -1;
     }
-    else {
+
+    if (id3v2_offset < 0) {
         char tag[3];
 #if 0
-        fprintf(stderr, "id3v2 tag not found in file %s. trying id3v1\n", finfo->path);
+        fprintf(stderr, "id3v2 tag not found in file %s. trying id3v1\n",
+                finfo->path);
 #endif
         /* check for id3v1 tag */
         if (lseek(fd, -128, SEEK_END) == -1) {
@@ -708,11 +694,6 @@ _parse(struct plugin *plugin, struct lms_context *ctxt, const struct lms_file_in
         }
     }
 
-    lms_string_size_strip_and_free(&info.title);
-    lms_string_size_strip_and_free(&info.artist);
-    lms_string_size_strip_and_free(&info.album);
-    lms_string_size_strip_and_free(&info.genre);
-
     if (!info.title.str) {
         int ext_idx;
         ext_idx = ((int)match) - 1;
@@ -741,6 +722,7 @@ _parse(struct plugin *plugin, struct lms_context *ctxt, const struct lms_file_in
     r = lms_db_audio_add(plugin->audio_db, &audio_info);
 
   done:
+    posix_fadvise(fd, 0, 0, POSIX_FADV_DONTNEED);
     close(fd);
 
     if (info.title.str)
